@@ -1,28 +1,34 @@
 '''
-adm.py - this file is part of S3QL (http://s3ql.googlecode.com)
+adm.py - this file is part of S3QL.
 
-Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>
 
-This program can be distributed under the terms of the GNU GPLv3.
+This work can be distributed under the terms of the GNU GPLv3.
 '''
 
 from .logging import logging, QuietError, setup_logging
 from . import CURRENT_FS_REV, REV_VER_MAP
-from .backends.common import BetterBackend, get_backend, DanglingStorageURLError
-from .common import (get_backend_cachedir, get_seq_no, stream_write_bz2,
-                     stream_read_bz2, PICKLE_PROTOCOL, is_mounted)
-from .metadata import restore_metadata, cycle_metadata, dump_metadata
+from .backends.comprenc import ComprencBackend
+from .backends.common import NoSuchObject
+from .database import Connection
+from .common import (get_seq_no, is_mounted, get_backend, load_params,
+                     save_params, handle_on_return, get_backend_factory,
+                     AsyncFn)
+from .metadata import dump_and_upload_metadata, download_metadata
 from .parse_args import ArgumentParser
 from datetime import datetime as Datetime
 from getpass import getpass
+from queue import Queue, Full as QueueFull
 import os
-import pickle
+import s3ql.backends.common
+import s3ql.backends.s3c
+import hashlib
+import hmac
+import struct
 import shutil
 import sys
-import tempfile
 import textwrap
 import time
-import atexit
 
 log = logging.getLogger(__name__)
 
@@ -44,23 +50,24 @@ def parse_args(args):
                                        help='may be either of')
     subparsers.add_parser("passphrase", help="change file system passphrase",
                           parents=[pparser])
-    subparsers.add_parser("upgrade", help="upgrade file system to newest revision",
-                          parents=[pparser])
     subparsers.add_parser("clear", help="delete file system and all data",
                           parents=[pparser])
     subparsers.add_parser("download-metadata",
                           help="Interactively download metadata backups. "
                                "Use only if you know what you are doing.",
                           parents=[pparser])
+    sparser = subparsers.add_parser("upgrade", help="upgrade file system to newest revision",
+                          parents=[pparser])
+    sparser.add_argument("--threads", type=int, default=20,
+                        help='Number of threads to use')
 
-    parser.add_debug_modules()
+    parser.add_debug()
     parser.add_quiet()
     parser.add_log()
     parser.add_authfile()
-    parser.add_ssl()
+    parser.add_backend_options()
     parser.add_cachedir()
     parser.add_version()
-    parser.add_fatal_warnings()
 
     options = parser.parse_args(args)
 
@@ -81,32 +88,19 @@ def main(args=None):
         raise QuietError('Can not work on mounted file system.')
 
     if options.action == 'clear':
-        try:
-            backend = get_backend(options, plain=True)
-            atexit.register(backend.close)
-        except DanglingStorageURLError as exc:
-            raise QuietError(str(exc))
-        return clear(backend,
-                     get_backend_cachedir(options.storage_url, options.cachedir))
+        with get_backend(options, raw=True) as backend:
+            return clear(backend, options)
+    elif options.action == 'upgrade':
+        return upgrade(options)
 
-    try:
-        backend = get_backend(options)
-        atexit.register(backend.close)
-    except DanglingStorageURLError as exc:
-        raise QuietError(str(exc))
+    with get_backend(options) as backend:
+        if options.action == 'passphrase':
+            return change_passphrase(backend)
 
-    if options.action == 'upgrade':
-        return upgrade(backend, get_backend_cachedir(options.storage_url,
-                                                   options.cachedir))
+        elif options.action == 'download-metadata':
+            return download_metadata_cmd(backend, options)
 
-    if options.action == 'passphrase':
-        return change_passphrase(backend)
-
-    if options.action == 'download-metadata':
-        return download_metadata(backend, options.storage_url)
-
-
-def download_metadata(backend, storage_url):
+def download_metadata_cmd(backend, options):
     '''Download old metadata backups'''
 
     backups = sorted(backend.list('s3ql_metadata'))
@@ -122,7 +116,7 @@ def download_metadata(backend, storage_url):
         except:
             log.error('Error retrieving information about %s, skipping', name)
             continue
-        
+
         if 'last-modified' in params:
             date = Datetime.fromtimestamp(params['last-modified']).strftime('%Y-%m-%d %H:%M:%S')
         else:
@@ -139,36 +133,24 @@ def download_metadata(backend, storage_url):
         except:
             log.warning('Invalid input')
 
-    cachepath = get_backend_cachedir(storage_url, '.')
+    cachepath = options.cachepath
     for i in ('.db', '.params'):
         if os.path.exists(cachepath + i):
             raise QuietError('%s already exists, aborting.' % cachepath + i)
 
     param = backend.lookup(name)
-    with tempfile.TemporaryFile() as tmpfh:
-        def do_read(fh):
-            tmpfh.seek(0)
-            tmpfh.truncate()
-            stream_read_bz2(fh, tmpfh)
-            
-        log.info('Downloading and decompressing %s...', name)
-        backend.perform_read(do_read, name)
-
-        log.info("Reading metadata...")
-        tmpfh.seek(0)
-        restore_metadata(tmpfh, cachepath + '.db')
+    download_metadata(backend, cachepath + ".db", name)
 
     # Raise sequence number so that fsck.s3ql actually uses the
     # downloaded backup
     seq_nos = [ int(x[len('s3ql_seq_no_'):]) for x in backend.list('s3ql_seq_no_') ]
     param['seq_no'] = max(seq_nos) + 1
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
+    save_params(cachepath, param)
 
 def change_passphrase(backend):
     '''Change file system passphrase'''
 
-    if not isinstance(backend, BetterBackend) and backend.passphrase:
+    if not isinstance(backend, ComprencBackend) and backend.passphrase:
         raise QuietError('File system is not encrypted.')
 
     data_pw = backend.passphrase
@@ -188,24 +170,22 @@ def change_passphrase(backend):
     backend['s3ql_passphrase_bak3'] = data_pw
     backend.passphrase = data_pw
 
-    
-def clear(backend, cachepath):
+def clear(backend, options):
     print('I am about to delete all data in %s.' % backend,
           'This includes any S3QL file systems as well as any other stored objects.',
           'Please enter "yes" to continue.', '> ', sep='\n', end='')
     sys.stdout.flush()
-    
+
     if sys.stdin.readline().strip().lower() != 'yes':
         raise QuietError()
 
     log.info('Deleting...')
-
     for suffix in ('.db', '.params'):
-        name = cachepath + suffix
+        name = options.cachepath + suffix
         if os.path.exists(name):
             os.unlink(name)
 
-    name = cachepath + '-cache'
+    name = options.cachepath + '-cache'
     if os.path.exists(name):
         shutil.rmtree(name)
 
@@ -217,39 +197,35 @@ def clear(backend, cachepath):
 def get_old_rev_msg(rev, prog):
     return textwrap.dedent('''\
         The last S3QL version that supported this file system revision
-        was %(version)s. You can run this version's %(prog)s by executing:
-        
-          $ wget http://s3ql.googlecode.com/files/s3ql-%(version)s.tar.bz2
+        was %(version)s. To run this version's %(prog)s, proceed along
+        the following steps:
+
+          $ wget http://s3ql.googlecode.com/files/s3ql-%(version)s.tar.bz2 \
+            || wget https://bitbucket.org/nikratio/s3ql/downloads/s3ql-%(version)s.tar.bz2
           $ tar xjf s3ql-%(version)s.tar.bz2
-          $ (cd s3ql-%(version)s; ./setup.py build_ext)
+          $ (cd s3ql-%(version)s; ./setup.py build_ext --inplace)
           $ s3ql-%(version)s/bin/%(prog)s <options>
         ''' % { 'version': REV_VER_MAP[rev],
                 'prog': prog })
 
-def upgrade(backend, cachepath):
+@handle_on_return
+def upgrade(options, on_return):
     '''Upgrade file system to newest revision'''
 
+    s3ql.backends.comprenc.UPGRADE_MODE = checksum_basic_mapping_old
+    s3ql.backends.s3c.UPGRADE_MODE = checksum_basic_mapping_old
+
     log.info('Getting file system parameters..')
 
-    seq_nos = list(backend.list('s3ql_seq_no_'))
-    if (seq_nos[0].endswith('.meta')
-        or seq_nos[0].endswith('.dat')):
-        print(textwrap.dedent(''' 
-            File system revision too old to upgrade!
-            
-            You need to use an older S3QL version to upgrade to a more recent
-            revision before you can use this version to upgrade to the newest
-            revision.
-            '''))
-        print(get_old_rev_msg(11 + 1, 's3qladm'))
-        raise QuietError()
-    seq_no = get_seq_no(backend)
+    backend_factory = get_backend_factory(options)
+    backend = on_return.enter_context(backend_factory())
 
     # Check for cached metadata
+    cachepath = options.cachepath
     db = None
+    seq_no = get_seq_no(backend)
     if os.path.exists(cachepath + '.params'):
-        with open(cachepath + '.params', 'rb') as fh:
-            param = pickle.load(fh)
+        param = load_params(cachepath)
         if param['seq_no'] < seq_no:
             log.info('Ignoring locally cached metadata (outdated).')
             param = backend.lookup('s3ql_metadata')
@@ -258,12 +234,8 @@ def upgrade(backend, cachepath):
             print(get_old_rev_msg(param['revision'], 'fsck.s3ql'))
             raise QuietError()
         else:
-            # Normally we'd use cached metadata now, but for this upgrade
-            # we actually have to ensure that we restore the sqlite db
-            # from the metadata dump.
-            #log.info('Using cached metadata.')
-            #db = Connection(cachepath + '.db')
-            log.info('Ignoring locally cached metadata for upgrade.')
+            log.info('Using cached metadata.')
+            db = Connection(cachepath + '.db')
     else:
         param = backend.lookup('s3ql_metadata')
 
@@ -287,10 +259,10 @@ def upgrade(backend, cachepath):
         raise QuietError()
 
     # Check revision
-    if param['revision'] < 16:
-        print(textwrap.dedent(''' 
+    if param['revision'] < CURRENT_FS_REV-1:
+        print(textwrap.dedent('''
             File system revision too old to upgrade!
-            
+
             You need to use an older S3QL version to upgrade to a more recent
             revision before you can use this version to upgrade to the newest
             revision.
@@ -303,84 +275,177 @@ def upgrade(backend, cachepath):
         return
 
     print(textwrap.dedent('''
-        I am about to update the file system to the newest revision. 
+        I am about to update the file system to the newest revision.
         You will not be able to access the file system with any older version
-        of S3QL after this operation. 
-        
+        of S3QL after this operation.
+
         You should make very sure that this command is not interrupted and
         that no one else tries to mount, fsck or upgrade the file system at
-        the same time.
-        
+        the same time. You may interrupt the update with Ctrl+C and resume
+        at a later time, but the filesystem will not be in a usable state
+        until the upgrade is complete.
         '''))
 
     print('Please enter "yes" to continue.', '> ', sep='\n', end='')
     sys.stdout.flush()
-    
+
     if sys.stdin.readline().strip().lower() != 'yes':
         raise QuietError()
 
-    # For this upgrade, we just need to recreate the sqlite database from the
-    # metadata dump, because some SQLite types have changed
-    assert db is None
-    
-    # Keep backup of local metadata (just in case...)
-    if os.path.exists(cachepath + '.params'):
-        assert os.path.exists(cachepath + '.db')
-        if (os.path.exists(cachepath + '.db.bak') or
-            os.path.exists(cachepath + '.params.bak')):
-            raise QuietError('Metadata backup already exists, did something go wrong?')
-        os.rename(cachepath + '.db', cachepath + '.db.bak')
-        os.rename(cachepath + '.params', cachepath + '.params.bak')
-    
     if not db:
-        # Need to download metadata
-        with tempfile.TemporaryFile() as tmpfh:
-            def do_read(fh):
-                tmpfh.seek(0)
-                tmpfh.truncate()
-                stream_read_bz2(fh, tmpfh)
+        db = download_metadata(backend, cachepath + '.db')
 
-            log.info("Downloading & uncompressing metadata...")
-            backend.perform_read(do_read, "s3ql_metadata")
-
-            log.info("Reading metadata...")
-            tmpfh.seek(0)
-            db = restore_metadata(tmpfh, cachepath + '.db')
-
-            
     log.info('Upgrading from revision %d to %d...', param['revision'], CURRENT_FS_REV)
 
+    # Upgrade all objects, so that we can remove legacy conversion
+    # routines in the next release.
+    if backend.passphrase:
+        update_obj_metadata(backend, backend_factory, db, options)
+
     param['revision'] = CURRENT_FS_REV
     param['last-modified'] = time.time()
     param['seq_no'] += 1
 
-    log.info('Dumping metadata...')
-    with tempfile.TemporaryFile() as fh:
-        dump_metadata(db, fh)
-        def do_write(obj_fh):
-            fh.seek(0)
-            stream_write_bz2(fh, obj_fh)
-            return obj_fh
-
-        log.info("Compressing and uploading metadata...")
-        backend.store('s3ql_seq_no_%d' % param['seq_no'], b'Empty')
-        obj_fh = backend.perform_write(do_write, "s3ql_metadata_new", metadata=param,
-                                      is_compressed=True)
-        
-    log.info('Wrote %.2f MiB of compressed metadata.', obj_fh.get_obj_size() / 1024 ** 2)
-    log.info('Cycling metadata backups...')
-    cycle_metadata(backend)
-    
+    dump_and_upload_metadata(backend, db, param)
     backend['s3ql_seq_no_%d' % param['seq_no']] = b'Empty'
 
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
+    print('File system upgrade complete.')
 
-    log.info('Cleaning up local metadata...')
-    db.execute('ANALYZE')
-    db.execute('VACUUM')
 
+def update_obj_metadata(backend, backend_factory, db, options):
+    '''Upgrade metadata of storage objects'''
+
+    # No need to update sequence number, since we are going to
+    # write out a new one after the upgrade.
+    extra_objects = { 's3ql_metadata',
+                      's3ql_passphrase', 's3ql_passphrase_bak1',
+                      's3ql_passphrase_bak2', 's3ql_passphrase_bak3' }
+    extra_objects |= { 's3ql_metadata_bak_%d' % i for i in range(30) }
+
+    def yield_objects():
+        for obj_id in extra_objects:
+            yield obj_id
+        for (id_,) in db.query('SELECT id FROM objects'):
+            yield 's3ql_data_%d' % id_
+    total = db.get_val('SELECT COUNT(id) FROM objects') + len(extra_objects)
+
+    queue = Queue(maxsize=options.threads)
+    threads = []
+    for _ in range(options.threads):
+        t = AsyncFn(upgrade_loop, queue, options, backend_factory)
+        # Don't wait for worker threads, gives deadlock if main thread
+        # terminates with exception
+        t.daemon = True
+        t.start()
+        threads.append(t)
+
+    stamp = time.time()
+    for (i, obj_id) in enumerate(yield_objects()):
+        stamp2 = time.time()
+        if stamp2 - stamp > 1:
+            sys.stdout.write('\r..processed %d/%d objects (%d%%)..'
+                             % (i, total, i/total*100))
+            sys.stdout.flush()
+            stamp = stamp2
+
+            # Terminate early if any thread failed with an exception
+            for t in threads:
+                if not t.is_alive():
+                    t.join_and_raise()
+
+        # Avoid blocking if all threads terminated
+        while True:
+            try:
+                queue.put(obj_id, timeout=1)
+            except QueueFull:
+                pass
+            else:
+                break
+            for t in threads:
+                if not t.is_alive():
+                    t.join_and_raise()
+
+    queue.maxsize += len(threads)
+    for t in threads:
+        queue.put(None)
+
+    for t in threads:
+        t.join_and_raise()
+
+    sys.stdout.write('\n')
+
+def upgrade_loop(queue, options, backend_factory):
+
+    with backend_factory() as backend:
+        while True:
+            key = queue.get()
+            if key is None:
+                break
+
+            # When reading passphrase objects, we have to use the
+            # "outer" password
+            if key.startswith('s3ql_passphrase'):
+                data_pw = backend.passphrase
+                backend.passphrase = options.fs_passphrase
+            try:
+                try:
+                    meta = backend.lookup(key)
+                except NoSuchObject:
+                    continue
+
+                if meta['needs_reupload']:
+                    del meta['needs_reupload']
+                    backend.update_meta(key, meta)
+            finally:
+                if key.startswith('s3ql_passphrase'):
+                    backend.passphrase = data_pw
+
+def checksum_basic_mapping_old(metadata, key=None):
+    '''Compute checksum for mapping of elementary types
+
+    Keys of *d* must be strings. Values of *d* must be of elementary
+    type (i.e., `str`, `bytes`, `int`, `float`, `complex`, `bool` or
+    None). If there is a key named ``signature``, then it is excluded
+    from the checksum computation.
+
+    If *key* is None, compute MD5. Otherwise compute HMAC using *key*.
+    '''
+
+    # In order to compute a safe checksum, we need to convert each object to a
+    # unique representation (i.e, we can't use repr(), as it's output may change
+    # in the future). Furthermore, we have to make sure that the representation
+    # is theoretically reversible, or there is a potential for collision
+    # attacks.
+
+    if key is None:
+        chk = hashlib.md5()
+    else:
+        chk = hmac.new(key, digestmod=hashlib.sha256)
+
+    for mkey in sorted(metadata.keys()):
+        assert isinstance(mkey, str)
+        if mkey == 'signature':
+            continue
+        chk.update(mkey.encode('utf-8'))
+        val = metadata[mkey]
+        if isinstance(val, str):
+            val = b'\0s' + val.encode('utf-8') + b'\0'
+        elif val is None:
+            val = b'\0n'
+        elif isinstance(val, int):
+            val = b'\0i' + ('%d' % val).encode() + b'\0'
+        elif isinstance(val, bool):
+            val = b'\0t' if val else b'\0f'
+        elif isinstance(val, float):
+            val = b'\0d' + struct.pack('<d', val)
+        elif isinstance(val, (bytes, bytearray)):
+            assert len(val).bit_length() <= 32
+            val = b'\0b' + struct.pack('<I', len(val))
+        else:
+            raise ValueError("Don't know how to checksum %s instances" % type(val))
+        chk.update(val)
+
+    return chk.digest()
 
 if __name__ == '__main__':
     main(sys.argv[1:])
-
