@@ -1,21 +1,21 @@
 '''
-mount.py - this file is part of S3QL (http://s3ql.googlecode.com)
+mount.py - this file is part of S3QL.
 
-Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>
 
-This program can be distributed under the terms of the GNU GPLv3.
+This work can be distributed under the terms of the GNU GPLv3.
 '''
 
 from .logging import logging, setup_logging, QuietError
 from . import fs, CURRENT_FS_REV
-from .backends.common import get_backend_factory, BackendPool, DanglingStorageURLError
+from .backends.pool import BackendPool
 from .block_cache import BlockCache
-from .common import (get_backend_cachedir, get_seq_no, stream_write_bz2, stream_read_bz2,
-                     PICKLE_PROTOCOL)
+from .common import (get_seq_no, get_backend_factory, load_params, save_params)
 from .daemonize import daemonize
 from .database import Connection
 from .inode_cache import InodeCache
-from .metadata import cycle_metadata, dump_metadata, restore_metadata
+from .metadata import (download_metadata, upload_metadata, dump_and_upload_metadata,
+                       dump_metadata)
 from .parse_args import ArgumentParser
 from .exit_stack import ExitStack
 from threading import Thread
@@ -23,10 +23,10 @@ import _thread
 import argparse
 import faulthandler
 import llfuse
+import itertools
 import os
 import platform
 import subprocess
-import pickle
 import re
 import signal
 import resource
@@ -37,11 +37,16 @@ import time
 import shutil
 import atexit
 
+try:
+    from systemd.daemon import notify as sd_notify
+except ImportError:
+    sd_notify = None
+
 log = logging.getLogger(__name__)
 
 def install_thread_excepthook():
     """work around sys.excepthook thread bug
-    
+
     See http://bugs.python.org/issue1230540.
 
     Call once from __main__ before creating any threads. If using
@@ -72,13 +77,13 @@ def main(args=None):
         args = sys.argv[1:]
 
     options = parse_args(args)
-    
+
     # Save handler so that we can remove it when daemonizing
     stdout_log_handler = setup_logging(options)
 
     if not os.path.exists(options.mountpoint):
-        raise QuietError('Mountpoint does not exist.')
-        
+        raise QuietError('Mountpoint does not exist.', exitcode=36)
+
     if options.threads is None:
         options.threads = determine_threads(options)
 
@@ -91,10 +96,11 @@ def main(args=None):
     # each upload thread (because each thread is using at least one socket and
     # at least one temporary file)
     avail_fd -= 32 + 3 * options.threads
-    
+
     if options.max_cache_entries is None:
         if avail_fd <= 64:
-            raise QuietError("Not enough available file descriptors.")
+            raise QuietError("Not enough available file descriptors.",
+                             exitcode=37)
         log.info('Autodetected %d file descriptors available for cache entries',
                  avail_fd)
         options.max_cache_entries = avail_fd
@@ -112,19 +118,17 @@ def main(args=None):
     backend_factory = get_backend_factory(options)
     backend_pool = BackendPool(backend_factory)
     atexit.register(backend_pool.flush)
-    
+
     # Get paths
-    cachepath = get_backend_cachedir(options.storage_url, options.cachedir)
+    cachepath = options.cachepath
 
     # Retrieve metadata
-    try:
-        with backend_pool() as backend:
-            (param, db) = get_metadata(backend, cachepath)
-    except DanglingStorageURLError as exc:
-        raise QuietError(str(exc))
+    with backend_pool() as backend:
+        (param, db) = get_metadata(backend, cachepath)
 
-    if param['max_obj_size'] < options.min_obj_size:
-        raise QuietError('Maximum object size must be bigger than minimum object size.')
+    #if param['max_obj_size'] < options.min_obj_size:
+    #    raise QuietError('Maximum object size must be bigger than minimum object size.',
+    #                     exitcode=2)
 
     # Handle --cachesize
     rec_cachesize = options.max_cache_entries * param['max_obj_size'] / 2
@@ -133,9 +137,9 @@ def main(args=None):
         options.cachesize = min(rec_cachesize, 0.8 * avail_cache)
         log.info('Setting cache size to %d MB', options.cachesize / 1024)
     elif options.cachesize > avail_cache:
-        log.warning('Warning! Requested cache size %d MB, but only %d MB available',
+        log.warning('Requested cache size %d MB, but only %d MB available',
                     options.cachesize / 1024, avail_cache / 1024)
-        
+
     if options.nfs:
         # NFS may try to look up '..', so we have to speed up this kind of query
         log.info('Creating NFS indices...')
@@ -152,38 +156,39 @@ def main(args=None):
     operations = fs.Operations(block_cache, db, max_obj_size=param['max_obj_size'],
                                inode_cache=InodeCache(db, param['inode_gen']),
                                upload_event=metadata_upload_thread.event)
+    block_cache.fs = operations
+    metadata_upload_thread.fs = operations
 
     with ExitStack() as cm:
-        log.info('Mounting filesystem...')
+        log.info('Mounting %s at %s...', options.storage_url, options.mountpoint)
         try:
             llfuse.init(operations, options.mountpoint, get_fuse_opts(options))
         except RuntimeError as exc:
-            raise QuietError(str(exc))
-        
+            raise QuietError(str(exc), exitcode=39)
+
         unmount_clean = False
         def unmount():
             log.info("Unmounting file system...")
-            # Acquire lock so that Operations.destroy() is called with the
-            # global lock like all other handlers
-            with llfuse.lock:
-                llfuse.close(unmount=unmount_clean)
+            llfuse.close(unmount=unmount_clean)
         cm.callback(unmount)
-        
+
         if options.fg:
             faulthandler.enable()
+            faulthandler.register(signal.SIGUSR1)
         else:
             if stdout_log_handler:
                 logging.getLogger().removeHandler(stdout_log_handler)
             global crit_log_fh
             crit_log_fh = open(os.path.join(options.cachedir, 'mount.s3ql_crit.log'), 'a')
             faulthandler.enable(crit_log_fh)
+            faulthandler.register(signal.SIGUSR1, file=crit_log_fh)
             daemonize(options.cachedir)
 
         mark_metadata_dirty(backend, cachepath, param)
-        
+
         block_cache.init(options.threads)
         cm.callback(block_cache.destroy)
-        
+
         metadata_upload_thread.start()
         cm.callback(metadata_upload_thread.join)
         cm.callback(metadata_upload_thread.stop)
@@ -191,15 +196,23 @@ def main(args=None):
         commit_thread.start()
         cm.callback(commit_thread.join)
         cm.callback(commit_thread.stop)
-        
+
         if options.upstart:
             os.kill(os.getpid(), signal.SIGSTOP)
+        if sd_notify is not None:
+            sd_notify('READY=1')
+            sd_notify('MAINPID=%d' % os.getpid())
 
         exc_info = setup_exchook()
+        workers = 1 if options.single else None # use default
+
         if options.profile:
-            prof.runcall(llfuse.main, options.single)
+            ret = prof.runcall(llfuse.main, workers)
         else:
-            llfuse.main(options.single)
+            ret = llfuse.main(workers)
+
+        if ret is not None:
+            raise RuntimeError('Received signal %d, terminating' % (ret,))
 
         # Allow operations to terminate while block_cache is still available
         # (destroy() will be called again when from llfuse.close(), but at that
@@ -230,27 +243,11 @@ def main(args=None):
             log.info('File system unchanged, not uploading metadata.')
             del backend['s3ql_seq_no_%d' % param['seq_no']]
             param['seq_no'] -= 1
-            with open(cachepath + '.params', 'wb') as fh:
-                pickle.dump(param, fh, PICKLE_PROTOCOL)
+            save_params(cachepath, param)
         elif seq_no == param['seq_no']:
             param['last-modified'] = time.time()
-
-            log.info('Dumping metadata...')
-            with tempfile.TemporaryFile() as fh:
-                dump_metadata(db, fh)
-                def do_write(obj_fh):
-                    fh.seek(0)
-                    stream_write_bz2(fh, obj_fh)
-                    return obj_fh
-
-                log.info("Compressing and uploading metadata...")
-                obj_fh = backend.perform_write(do_write, "s3ql_metadata_new",
-                                               metadata=param, is_compressed=True)
-            log.info('Wrote %.2f MiB of compressed metadata', obj_fh.get_obj_size() / 1024 ** 2)
-            log.info('Cycling metadata backups...')
-            cycle_metadata(backend)
-            with open(cachepath + '.params', 'wb') as fh:
-                pickle.dump(param, fh, PICKLE_PROTOCOL)
+            dump_and_upload_metadata(backend, db, param)
+            save_params(cachepath, param)
         else:
             log.error('Remote metadata is newer than local (%d vs %d), '
                       'refusing to overwrite!', seq_no, param['seq_no'])
@@ -287,7 +284,7 @@ def get_system_memory():
     If amount cannot be determined, emits warning and
     returns -1.
     '''
-    
+
     # MacOS X doesn't support sysconf('SC_PHYS_PAGES')
     if platform.system() == 'Darwin':
         try:
@@ -303,9 +300,9 @@ def get_system_memory():
         if not hit:
             log.warning('Cannot determine system memory, unable to parse sysctl output.')
             return -1
-            
+
         return int(hit.group(1))
-        
+
     else:
         try:
             return os.sysconf('SC_PHYS_PAGES') * os.sysconf('SC_PAGESIZE')
@@ -313,7 +310,7 @@ def get_system_memory():
             log.warning('Unable to determine number of CPU cores (sysconf failed).')
             return -1
 
-            
+
 # Memory required for LZMA compression in MB (from xz(1))
 LZMA_MEMORY = { 0: 3, 1: 9, 2: 17, 3: 32, 4: 48,
                 5: 94, 6: 94, 7: 186, 8: 370, 9: 674 }
@@ -347,8 +344,8 @@ def determine_threads(options):
         if threads > 0:
             log.info('Using %d upload threads (memory limited).', threads)
         else:
-            log.warning('Warning: compression will require %d MiB memory '
-                     '(%d%% of total system memory', mem_per_thread / 1024 ** 2,
+            log.warning('Compression will require %d MiB memory '
+                        '(%d%% of total system memory', mem_per_thread / 1024 ** 2,
                      mem_per_thread * 100 / memory)
             threads = 1
         return threads
@@ -362,82 +359,89 @@ def get_metadata(backend, cachepath):
 
     seq_no = get_seq_no(backend)
 
+    # When there was a crash during metadata rotation, we may end up
+    # without an s3ql_metadata object.
+    meta_obj_name = 's3ql_metadata'
+    if meta_obj_name not in backend:
+        meta_obj_name += '_new'
+
     # Check for cached metadata
     db = None
     if os.path.exists(cachepath + '.params'):
-        with open(cachepath + '.params', 'rb') as fh:
-            param = pickle.load(fh)
+        param = load_params(cachepath)
         if param['seq_no'] < seq_no:
             log.info('Ignoring locally cached metadata (outdated).')
-            param = backend.lookup('s3ql_metadata')
+            param = backend.lookup(meta_obj_name)
         elif param['seq_no'] > seq_no:
-            raise QuietError("File system not unmounted cleanly, run fsck!")            
+            raise QuietError("File system not unmounted cleanly, run fsck!",
+                             exitcode=30)
         else:
             log.info('Using cached metadata.')
             db = Connection(cachepath + '.db')
     else:
-        param = backend.lookup('s3ql_metadata')
+        param = backend.lookup(meta_obj_name)
 
     # Check for unclean shutdown
     if param['seq_no'] < seq_no:
-        raise QuietError('Backend reports that fs is still mounted elsewhere, aborting.')       
+        raise QuietError('Backend reports that fs is still mounted elsewhere, aborting.',
+                         exitcode=31)
 
     # Check revision
     if param['revision'] < CURRENT_FS_REV:
-        raise QuietError('File system revision too old, please run `s3qladm upgrade` first.')
+        raise QuietError('File system revision too old, please run `s3qladm upgrade` first.',
+                         exitcode=32)
     elif param['revision'] > CURRENT_FS_REV:
         raise QuietError('File system revision too new, please update your '
-                         'S3QL installation.')
+                         'S3QL installation.', exitcode=33)
 
     # Check that the fs itself is clean
     if param['needs_fsck']:
-        raise QuietError("File system damaged or not unmounted cleanly, run fsck!")
+        raise QuietError("File system damaged or not unmounted cleanly, run fsck!",
+                         exitcode=30)
     if time.time() - param['last_fsck'] > 60 * 60 * 24 * 31:
         log.warning('Last file system check was more than 1 month ago, '
                  'running fsck.s3ql is recommended.')
 
     if  param['max_inode'] > 2 ** 32 - 50000:
-        raise QuietError('Insufficient free inodes, fsck run required.')
+        raise QuietError('Insufficient free inodes, fsck run required.',
+                         exitcode=34)
     elif param['max_inode'] > 2 ** 31:
         log.warning('Few free inodes remaining, running fsck is recommended')
 
+    if os.path.exists(cachepath + '-cache'):
+        for i in itertools.count():
+            bak_name = '%s-cache.bak%d' % (cachepath, i)
+            if not os.path.exists(bak_name):
+                break
+        log.warning('Found outdated cache directory (%s), renaming to .bak%d',
+                    cachepath + '-cache', i)
+        log.warning('You should delete this directory once you are sure that '
+                    'everything is in order.')
+        os.rename(cachepath + '-cache', bak_name)
+
     # Download metadata
     if not db:
-        with tempfile.TemporaryFile() as tmpfh:
-            def do_read(fh):
-                tmpfh.seek(0)
-                tmpfh.truncate()
-                stream_read_bz2(fh, tmpfh)
+        db = download_metadata(backend, cachepath + '.db')
+    save_params(cachepath, param)
 
-            log.info('Downloading and decompressing metadata...')
-            backend.perform_read(do_read, "s3ql_metadata")
-            
-            log.info("Reading metadata...")
-            tmpfh.seek(0)
-            db = restore_metadata(tmpfh, cachepath + '.db')
-
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
-            
     return (param, db)
 
 def mark_metadata_dirty(backend, cachepath, param):
     '''Mark metadata as dirty and increase sequence number'''
-    
+
     param['seq_no'] += 1
     param['needs_fsck'] = True
+    save_params(cachepath, param)
     backend['s3ql_seq_no_%d' % param['seq_no']] = b'Empty'
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
     param['needs_fsck'] = False
 
-
 def get_fuse_opts(options):
     '''Return fuse options for given command line options'''
 
     fuse_opts = [ "nonempty", 'fsname=%s' % options.storage_url,
-                  'subtype=s3ql' ]
-    
+                  'subtype=s3ql', 'big_writes', 'max_write=131072',
+                  'no_remote_lock' ]
+
     if platform.system() == 'Darwin':
         # FUSE4X and OSXFUSE claim to support nonempty, but
         # neither of them actually do.
@@ -478,7 +482,8 @@ def parse_args(args):
                            'owner', 'users', 'nobootwait'):
                     continue
                 elif opt == 'ro':
-                    raise QuietError('Read-only mounting not supported.')
+                    raise QuietError('Read-only mounting not supported.',
+                                     exitcode=35)
                 args.insert(pos, '--' + opt)
 
     def compression_type(s):
@@ -496,19 +501,18 @@ def parse_args(args):
         if alg == 'none':
             alg =  None
         return (alg, lvl)
-        
+
     parser = ArgumentParser(
         description="Mount an S3QL file system.")
 
     parser.add_log('~/.s3ql/mount.log')
     parser.add_cachedir()
     parser.add_authfile()
-    parser.add_debug_modules()
+    parser.add_debug()
     parser.add_quiet()
-    parser.add_ssl()
+    parser.add_backend_options()
     parser.add_version()
     parser.add_storage_url()
-    parser.add_fatal_warnings()
 
     parser.add_argument("mountpoint", metavar='<mountpoint>', type=os.path.abspath,
                         help='Where to mount the file system')
@@ -520,11 +524,6 @@ def parse_args(args):
                       'this number you have to make sure that your process file descriptor '
                       'limit (as set with `ulimit -n`) is high enough (at least the number '
                       'of cache entries + 100).')
-    parser.add_argument("--min-obj-size", type=int, default=512, metavar='<size>',
-                        help=argparse.SUPPRESS)
-#                      help="Minimum size of storage objects in KiB. Files smaller than this "
-#                           "may be combined into groups that are stored as single objects "
-#                           "in the storage backend. Default: %(default)d KB.")
     parser.add_argument("--allow-other", action="store_true", default=False, help=
                       'Normally, only the user who called `mount.s3ql` can access the mount '
                       'point. This user then also has full access to it, independent of '
@@ -536,15 +535,9 @@ def parse_args(args):
                            'user and the root user.')
     parser.add_argument("--fg", action="store_true", default=False,
                       help="Do not daemonize, stay in foreground")
-    parser.add_argument("--single", action="store_true", default=False,
-                      help="Run in single threaded mode. If you don't understand this, "
-                           "then you don't need it.")
     parser.add_argument("--upstart", action="store_true", default=False,
                       help="Stay in foreground and raise SIGSTOP once mountpoint "
                            "is up.")
-    parser.add_argument("--profile", action="store_true", default=False,
-                      help="Create profiling information. If you don't understand this, "
-                           "then you don't need it.")
     parser.add_argument("--compress", action="store", default='lzma-6',
                         metavar='<algorithm-lvl>', type=compression_type,
                         help="Compression algorithm and compression level to use when "
@@ -561,6 +554,17 @@ def parse_args(args):
     parser.add_argument("--nfs", action="store_true", default=False,
                       help='Enable some optimizations for exporting the file system '
                            'over NFS. (default: %(default)s)')
+    parser.add_argument("--single", action="store_true", default=False,
+                        help=argparse.SUPPRESS)
+    parser.add_argument("--profile", action="store_true", default=False,
+                        help=argparse.SUPPRESS)
+
+    # Not yet implemented. When implementing this, don't forget to
+    # uncomment check against param['max_obj_size'] in main().
+    #parser.add_argument("--min-obj-size", type=int, default=512, metavar='<size>',
+    #                    help="Minimum size of storage objects in KiB. Files smaller than this "
+    #                    "may be combined into groups that are stored as single objects "
+    #                    "in the storage backend. Default: %(default)d KB.")
 
     options = parser.parse_args(args)
 
@@ -586,9 +590,9 @@ class MetadataUploadThread(Thread):
     Periodically upload metadata. Upload is done every `interval`
     seconds, and whenever `event` is set. To terminate thread,
     set `quit` attribute as well as `event` event.
-    
+
     This class uses the llfuse global lock. When calling objects
-    passed in the constructor, the global lock is acquired first.    
+    passed in the constructor, the global lock is acquired first.
     '''
 
     def __init__(self, backend_pool, param, db, interval):
@@ -603,8 +607,14 @@ class MetadataUploadThread(Thread):
         self.quit = False
         self.name = 'Metadata-Upload-Thread'
 
+        # Can't assign in constructor, because Operations instance needs
+        # access to self.event as well.
+        self.fs = None
+
     def run(self):
-        log.debug('MetadataUploadThread: start')
+        log.debug('started')
+
+        assert self.fs is not None
 
         while not self.quit:
             self.event.wait(self.interval)
@@ -625,48 +635,42 @@ class MetadataUploadThread(Thread):
                 fh = tempfile.TemporaryFile()
                 dump_metadata(self.db, fh)
 
-
             with self.backend_pool() as backend:
                 seq_no = get_seq_no(backend)
-                if seq_no != self.param['seq_no']:
+                if seq_no > self.param['seq_no']:
                     log.error('Remote metadata is newer than local (%d vs %d), '
-                              'refusing to overwrite!', seq_no, self.param['seq_no'])
+                              'refusing to overwrite and switching to failsafe mode!',
+                              seq_no, self.param['seq_no'])
+                    self.fs.failsafe = True
                     fh.close()
-                    continue
+                    break
 
                 fh.seek(0)
                 self.param['last-modified'] = time.time()
 
                 # Temporarily decrease sequence no, this is not the final upload
                 self.param['seq_no'] -= 1
-                def do_write(obj_fh):
-                    fh.seek(0)
-                    stream_write_bz2(fh, obj_fh)
-                    return obj_fh
-                log.info("Compressing and uploading metadata...")
-                obj_fh = backend.perform_write(do_write, "s3ql_metadata_new",
-                                               metadata=self.param, is_compressed=True)
-                log.info('Wrote %.2f MiB of compressed metadata.',
-                         obj_fh.get_obj_size() / 1024 ** 2)
-                log.info('Cycling metadata backups...')
-                cycle_metadata(backend)
+                upload_metadata(backend, fh, self.param)
                 self.param['seq_no'] += 1
 
                 fh.close()
                 self.db_mtime = new_mtime
 
-        log.debug('MetadataUploadThread: end')
+        # Break reference loop
+        self.fs = None
+
+        log.debug('finished')
 
     def stop(self):
         '''Signal thread to terminate'''
 
-        log.debug('MetadataUploadThread: stop() called')
+        log.debug('started')
         self.quit = True
         self.event.set()
 
 def setup_exchook():
     '''Send SIGTERM if any other thread terminates with an exception
-    
+
     The exc_info will be saved in the list object returned
     by this function.
     '''
@@ -703,7 +707,7 @@ def setup_exchook():
 class CommitThread(Thread):
     '''
     Periodically upload dirty blocks.
-    
+
     This class uses the llfuse global lock. When calling objects
     passed in the constructor, the global lock is acquired first.
     '''
@@ -716,11 +720,11 @@ class CommitThread(Thread):
         self.name = 'CommitThread'
 
     def run(self):
-        log.debug('CommitThread: start')
+        log.debug('started')
 
         while not self.stop_event.is_set():
             did_sth = False
-            
+
             with llfuse.lock:
                 stamp = time.time()
                 # Need to make copy, since we aren't allowed to change
@@ -740,23 +744,19 @@ class CommitThread(Thread):
                         break
                     if stamp - el.last_access < 10:
                         break
-                    if not el.dirty or el in self.block_cache.in_transit:
-                        continue
-
-                    self.block_cache.upload(el)
-                    did_sth = True
+                    if self.block_cache.upload_if_dirty(el):
+                        did_sth = True
 
             if not did_sth:
                 self.stop_event.wait(5)
 
-        log.debug('CommitThread: end')
+        log.debug('finished')
 
     def stop(self):
         '''Signal thread to terminate'''
 
-        log.debug('CommitThread: stop() called')
+        log.debug('started')
         self.stop_event.set()
 
-
 if __name__ == '__main__':
     main(sys.argv[1:])
